STYLE_KEY = "style"
STYLE_LATENTS_KEY = "style_latents"
CONTENT_LATENT_KEY = "content_latent"
STYLIZED_STYLE_LATENTS_KEY = "stylized_style_latents"
STYLIZED_CONTENT_LATENT_KEY = "stylized_content_latent"

INPUT_B_KEY = "input_b"
LABEL_B_KEY = "labels_b"
